// (C) 2010 University of Washington.
// The software is subject to an academic license.
// Terms are contained in the LICENSE file.
//
// Runtime: common scheduler stuff
//

#include "dmp-internal.h"
#include "dmp-internal-wb.h"

//--------------------------------------------------------------
// Scheduling barriers
//--------------------------------------------------------------

// We have two kinds of barriers: the inner-round barriers and the
// end-of-round barrier.  For a sequence of modes A|B|S, where 'S'
// is serial mode, we use two inner-round barriers and one end-of-round
// barriers, as in:
//
// inner-round
//   A -> B  ::  b1--; wait until (b1 == 0)
//   B -> S  ::  b2--; wait until (b2 == 0)
//
// end-of-round
//   S -> A  ::  reset b1; reset b2; signal b3

#ifdef DMP_ENABLE_MODEL_O_B_S
atomic_uint_t DMPbufferingBarrier; // inner-round barrier: * -> B
#endif
#ifdef DMP_ENABLE_BUFFERED_MODE
atomic_uint_t DMPcommitBarrier;    // inner-round barrier: B -> Bcommit
#endif
atomic_uint_t DMPserialBarrier;    // inner-round barrier: * -> S
atomic_uint_t DMProundBarrier;     // end-of-round barrier: S -> *
atomic_int_t  DMPscheduledThread;  // token for serial mode

atomic_uint64_t DMProundNumber;    // round counter

//
// Bookends
//

static const DmpThreadState FirstStateInRound =
#if defined(DMP_ENABLE_MODEL_O_S) || defined(DMP_ENABLE_MODEL_O_B_S) ||\
    defined(DMP_ENABLE_MODEL_STM)
        RunOwnership;
#elif defined(DMP_ENABLE_MODEL_B_S) || defined(DMP_ENABLE_MODEL_OB_S)
        RunBuffered;
#endif

static const DmpThreadState LastStateInRound =
#if defined(DMP_ENABLE_MODEL_O_S) || defined(DMP_ENABLE_MODEL_O_B_S) ||\
    defined(DMP_ENABLE_MODEL_STM)
        WaitForOwnership;
#elif defined(DMP_ENABLE_MODEL_B_S) || defined(DMP_ENABLE_MODEL_OB_S)
        WaitForBuffered;
#endif

//-----------------------------------------------------------------------
// Scheduler API:
//
// DMP_waitForSerialMode()
//   * Blocks until the thread is scheduled in serial mode in the
//     current quantum.  When this returns, no other thread is running.
//   * REQUIRES: the thread is not sleeping
//
// DMP_waitForSerialModeOrFastHandoff()
//   * Blocks until the thread is scheduled in serial mode OR rescheduled
//     in ownership mode, which can happen when another thread performs a
//     "fast handoff" of a resource that this thread is waiting on.
//   * REQUIRES: the thread state is RunOwnership
//   * REQUIRES: DMPMAP->handoffSpinlock is held
//
// DMP_waitForNextQuantum()
//   * Blocks until the end of the current quantum.  When this returns,
//     many other threads may be running.
//   * REQUIRES: the thread is not sleeping
//
// DMP_waitForSignal(int* signal)
//   * Waits until '*signal' is true.  The thread is removed from the
//     runnable queue and sleeps until '*signal' is true.  On linux,
//     signalling uses a futex for efficiency.
//   * REQUIRES: the thread is not sleeping
//
// DMP_sleepAndTerminate()
//   * Blocks until the thread is scheduled in serial mode, and then
//     removes the thread from the runnable queue.  Sets 'exited = 1'
//     and 'state = Sleeping'.  If the thread was being joined, the
//     joiner is signaled.
//   * REQUIRES: the thread is not sleeping
//-----------------------------------------------------------------------

static inline void internal_waitAtBarrier(atomic_uint_t* b) {
  // Wait at an inner-round barrier.
  if (__sync_sub_and_fetch(b, 1) != 0) {
    while ((*b) != 0)
      YIELD();
  }
#ifdef DMP_ENABLE_ROUND_TIMING
  else {
#ifdef DMP_ENABLE_BUFFERED_MODE
    if (b == &DMPcommitBarrier) {
      DMProundTimeTransition(&DMPparallelModeTime, &DMPcommitModeTime);
    } else if (b == &DMPserialBarrier) {
      DMProundTimeTransition(&DMPcommitModeTime, &DMPserialModeTime);
    } else {
      assert(false);
    }
#else
    assert(b == &DMPserialBarrier);
    DMProundTimeTransition(&DMPparallelModeTime, &DMPserialModeTime);
#endif
  }
#endif
}

static inline void internal_waitForToken() {
#ifdef DMP_ENABLE_EMPTY_SERIAL_MODE
  // Optimization: don't wait for the serial token if we don't need to exec serially.
  if (DMPMAP->state == WaitForSerialToken && !DMPMAP->needSerial)
    return;
  if (DMPMAP->state == WaitForSerialToken) {
    DmpThreadInfo *other = __sync_val_compare_and_swap(&DMPMAP->notifyWhenWaitingForSerial, NULL, DMPMAP);
    if (other != NULL) {
      other->otherThreadWaitingForSerial = true;
    }
  }
#endif
  // Wait for the serial token.
  const int me = DMPMAP->threadID;
  while (DMPscheduledThread != me)
    YIELD();
}

#ifdef DMP_ENABLE_WB_HBSYNC

static inline void internal_updateGlobalSchedulingChunk() {
  // Called at the end of parallel mode.
  DMPglobalSchedulingChunks[DMPMAP->threadID].val = 0;
  DMPMAP->triggerGlobalSchedulingChunkUpdate = 0;
}

static inline void internal_resetGlobalSchedulingChunks() {
  // Called at the end of a round: reset to INT_MAX.
  for (int i = 0; i < DMPthreadInfosSize; ++i)
    DMPglobalSchedulingChunks[i].val = INT_MAX;
}

#else

static inline void internal_updateGlobalSchedulingChunk() {}
static inline void internal_resetGlobalSchedulingChunks() {}

#endif

//
// Wait for and execute write-buffer commit
//

#if defined(DMP_ENABLE_BUFFERED_MODE)

static inline void internal_executeBufferedCommit() {
  DMP_ASSERT(DMPMAP->state == RunBuffered);

  // Commit.
  DMPMAP_setState(WaitForCommit);
  DMPwbUpdateStats();
  internal_updateGlobalSchedulingChunk();
  internal_waitAtBarrier(&DMPcommitBarrier);

#if defined(DMP_ENABLE_WB_PARALLEL_COMMIT)
  __sync_synchronize();
  DMPMAP_setState(RunCommit);
  DMP_commitBufferedWrites();

#elif defined(DMP_ENABLE_WB_NONDET_COMMIT)
  __sync_synchronize();
  DMPMAP_setState(RunCommit);
  DMP_commitBufferedWrites();

#else  // serial commit
  internal_waitForToken();
  __sync_synchronize();
  DMPMAP_setState(RunCommit);
  DMP_commitBufferedWrites();
  __sync_synchronize();
  DMPscheduledThread = DMPMAP->nextRunnableID;
#endif
}

#endif

//
// Transition to Serial
//

#if defined(DMP_ENABLE_MODEL_O_S) || defined(DMP_ENABLE_MODEL_STM)

__attribute__((noinline))
static void internal_waitForSerialMode(const int islocked) {
  DMP_ASSERT(DMPMAP->state == RunOwnership);
  DMPMAP_setState(WaitForSerial);
  internal_updateGlobalSchedulingChunk();

#ifdef DMP_ENABLE_FAST_HANDOFF
  const int blockedOnResource = (DMPMAP->blockedOn != NULL);
  if (islocked) {
    DMP_SPINLOCK_UNLOCK(&DMPMAP->handoffSpinlock);
  }
#endif

  // Wait for all threads to block.
  if (__sync_sub_and_fetch(&DMPserialBarrier, 1) != 0) {
#ifdef DMP_ENABLE_FAST_HANDOFF
    if (blockedOnResource) {
      while (DMPserialBarrier != 0) {
        // Did a fast-handoff wake us up?
        if (*((volatile DmpThreadState*)&DMPMAP->state) == RunOwnership) {
          __sync_synchronize();
          return;
        }
        YIELD();
      }
    } else {
      while (DMPserialBarrier != 0)
        YIELD();
    }
#else
    while (DMPserialBarrier != 0)
      YIELD();
#endif
  }
#ifdef DMP_ENABLE_ROUND_TIMING
  else {
    DMProundTimeTransition(&DMPparallelModeTime, &DMPserialModeTime);
  }
#endif

  // Enter serial mode.
  DMPMAP_setState(WaitForSerialToken);
  internal_waitForToken();
#ifdef DMP_ENABLE_FAST_HANDOFF
  DMPMAP->blockedOn = NULL;
#endif
  DMPMAP_setState(RunSerial);
  __sync_synchronize();
}

#elif defined(DMP_ENABLE_MODEL_B_S) ||\
      defined(DMP_ENABLE_MODEL_OB_S)

__attribute__((noinline))
static void internal_waitForSerialMode(const int dummy) {
  DMP_ASSERT(DMPMAP->state == RunBuffered);

  // Commit.
  internal_executeBufferedCommit();

  // Enter serial mode.
  DMPMAP_setState(WaitForSerial);
  internal_waitAtBarrier(&DMPserialBarrier);
  DMPwbResetQuantum();

  // Wait for the serial token.
  DMPMAP_setState(WaitForSerialToken);
  internal_waitForToken();

  DMPMAP_setState(RunSerial);
  __sync_synchronize();
}

#elif defined(DMP_ENABLE_MODEL_O_B_S)

__attribute__((noinline))
static void internal_waitForSerialMode(const int dummy) {
  if (DMPMAP->state == RunOwnership)
    DMP_waitForBufferingMode();

  // Commit.
  internal_executeBufferedCommit();

  // Enter serial mode.
  DMPMAP_setState(WaitForSerial);
  internal_waitAtBarrier(&DMPserialBarrier);
  DMPwbResetQuantum();

  // Wait for the serial token.
  DMPMAP_setState(WaitForSerialToken);
  internal_waitForToken();

  DMPMAP_setState(RunSerial);
  __sync_synchronize();
}

void DMP_waitForBufferingMode() {
  if (DMPMAP->state != RunBuffered) {
    assert(DMPMAP->state == RunOwnership);
    DMPMAP_setState(WaitForBuffered);
    internal_waitAtBarrier(&DMPbufferingBarrier);
    DMPMAP_setState(RunBuffered);
  }
}

#endif

static void do_waitForSerialMode(const bool needSerial) {
  if (DMPMAP->state != RunSerial) {
#ifdef DMP_ENABLE_INSTRUMENT_WORK
    DMPMAP->toserial_total++;
#endif
#ifdef DMP_ENABLE_EMPTY_SERIAL_MODE
    DMPMAP->needSerial = needSerial;
#endif
    internal_waitForSerialMode(0);
  }
}

void DMP_waitForSerialMode() {
  do_waitForSerialMode(true);
}
void DMP_waitForSerialModeForEndOfQuantum() {
  do_waitForSerialMode(false);
}

//
// End-Of-Round
//

void DMP_resetRound(const int oldRoundBarrier) {
  DMProundNumber++;
  DMPscheduledThread = DMPfirstRunnableID;
  DMPserialBarrier = DMPnumRunnableThreads;
#ifdef DMP_ENABLE_BUFFERED_MODE
  DMPcommitBarrier = DMPnumRunnableThreads;
#endif
#ifdef DMP_ENABLE_MODEL_O_B_S
  DMPbufferingBarrier = DMPnumRunnableThreads;
#endif
#ifdef DMP_ENABLE_ROUND_TIMING
  DMProundTimeTransition(&DMPserialModeTime, &DMPparallelModeTime);
#endif
#ifdef DMP_ENABLE_MODEL_STM
  extern void DMPstmCheckConflicts();
  DMPstmCheckConflicts();
#endif
#ifdef DMP_ENABLE_WB_HBSYNC
#endif
  internal_resetGlobalSchedulingChunks();
  __sync_synchronize();
  DMProundBarrier = 1 - oldRoundBarrier;
}

#undef DMP_resetRound

//
// Quantum Boundaries
//

static inline void DMP_startNextQuantum() {
  DMPMAP->schedulingChunk = DMP_SCHEDULING_CHUNK_SIZE;
#ifdef DMP_ENABLE_EMPTY_SERIAL_MODE
  DMPMAP->needSerial = false;
  DMPMAP->notifyWhenWaitingForSerial = NULL;
#endif
  DMPMAP_setState(FirstStateInRound);
  __sync_synchronize();
}

static inline void DMP_arriveAtRoundBarrier(const int oldRoundBarrier) {
  // We have just ended our quantum.
  // Now we decide who gets the serial token next.
#ifndef DMP_ENABLE_EMPTY_SERIAL_MODE
  const bool haveToken = true;
  const bool isEndOfRound = DMPMAP->isLastRunnable;
  const int nextToGetToken = DMPMAP->nextRunnableID;
#else
  // "Empty serial mode" optimization.
  bool haveToken = true;
  bool isEndOfRound = false;
  int nextToGetToken = -1;

  if (!DMPMAP->needSerial) {
    // We didn't execute in serial mode, so we don't need the token.
    // But we may have it anyway (e.g. if we're the first scheduled thread).
    haveToken = (DMPMAP->threadID == DMPscheduledThread);
  }

  if (haveToken) {
    for (DmpThreadInfo *dmp = DMPMAP; ; dmp = dmp->nextRunnable) {
      if (dmp->isLastRunnable) {
        isEndOfRound = true;
        break;
      }
      if (dmp == DMPMAP)
        continue;
      if (dmp->needSerial) {
        nextToGetToken = dmp->threadID;
        break;
      }

      // This thread may not yet have finished parallel mode.
      // We need to wait for that to occur before continuing.
      DMPMAP->otherThreadWaitingForSerial = false;
      if (__sync_bool_compare_and_swap(&dmp->notifyWhenWaitingForSerial, NULL, DMPMAP)) {
        while (!DMPMAP->otherThreadWaitingForSerial)
          YIELD();
        __sync_synchronize();
      }
      // Now look again.
      if (VOLATILE_LOAD(dmp->needSerial)) {
        nextToGetToken = dmp->threadID;
        break;
      }
    }
  }
#endif

  const bool sleeping = (DMPMAP->state == Sleeping);

  // If sleeping or terminating, leave the runnable queue.
  if (sleeping) {
    DMPthread_removeFromRunnableQueue(DMPMAP);
  }

  // If holding the serial token, pass it on.
  if (haveToken) {
    if (isEndOfRound) {
      DMP_resetRound(oldRoundBarrier);
    } else {
      __sync_synchronize();
      DMPscheduledThread = nextToGetToken;
    }
  }

  // If running, wait for the next round.
  if (!sleeping && !isEndOfRound) {
    while (DMProundBarrier == oldRoundBarrier)
      YIELD();
  }
}

__attribute__((noinline))
void DMP_waitForNextQuantum() {
  // Special case to avoid a race: threads must remember the
  // 'roundBarrier' from the time they woke so they don't miss
  // the end-of-round.
  if (DMPMAP->state == JustWokeUp) {
    while (DMProundBarrier == DMPMAP->roundBarrierAtStart)
      YIELD();
    // Now we enter the first mode.
    DMP_startNextQuantum();
    return;
  }

  // Read this now, before ending parallel mode.
  // After parallel mode, if !DMPMAP->needSerial, we may race
  // with the thread that resets this end-of-round barrier.
  const int oldRoundBarrier = DMProundBarrier;

  // Make sure we're in serial mode.
  DMP_waitForSerialModeForEndOfQuantum();
  DMPMAP_setState(LastStateInRound);

  // Transition to next quantum.
  DMP_arriveAtRoundBarrier(oldRoundBarrier);
  DMP_startNextQuantum();

  // Were we canceled?
  if (DMPMAP->canceled) {
    pthread_exit(PTHREAD_CANCELED);
  }
}

void DMP_waitForSignal(volatile int* signal) {
  if (*signal) return;
  // NB: since we update the runnable queue, we must execute in serial mode.
  DMP_waitForSerialMode();
  if (*signal) return;

  // Sleep this thread.
  DMPMAP_setState(Sleeping);
  DMP_arriveAtRoundBarrier(DMProundBarrier);
  DMP_ASSERT(DMPscheduledThread != DMPMAP->threadID);

#ifdef __linux__
  WAIT_ON_FUTEX(signal, 0);
#else
  while ((*signal) == 0)
    YIELD();
#endif

  // We've been woken.
  DMP_ASSERT(*signal);
  DMP_ASSERT(DMPMAP->state == JustWokeUp);
  DMP_waitForNextQuantum();
}

void DMP_sleepAndTerminate() {
  // NB: since we update the runnable queue, we must execute in serial mode.
  DMP_waitForSerialMode();
  DMPMAP_setState(Sleeping);

  const int oldRoundBarrier = DMProundBarrier;

  // Wake the waiting joiner, if any.
  if (DMPMAP->joiner != NULL) {
    // NB: this adds 'joiner' to the *front* of the runnable queue.
    DMPthread_addToRunnableQueue(DMPMAP->joiner);
    DMPMAP->joiner->state = JustWokeUp;
    DMPMAP->joiner->roundBarrierAtStart = oldRoundBarrier;
    __sync_synchronize();
  }

  DMPMAP->exited = 1;

#ifdef __linux__
  if (DMPMAP->joiner != NULL) {
    futex((int*)&(DMPMAP->exited), FUTEX_WAKE, 1, NULL, NULL, 0);
  }
#endif

  // Terminate this thread.
  DMP_arriveAtRoundBarrier(oldRoundBarrier);
}

//
// Fast handoff
//
// Crazy story: with DMP_ENABLE_FAST_HANDOFF defined, lu simlarge with 4
// threads is 15 seconds slower.  Turns out, if all code blocks guarded
// by DMP_ENABLE_FAST_HANDOFF are commented out, EXCEPT for the below two
// functions, lu is still 15 seconds slower, even though these functions
// are never called!  Removing these functions eliminates those extra 15
// seconds.  It appears LLVM does really poor code layout :-(
//

#ifdef DMP_ENABLE_FAST_HANDOFF

void DMP_waitForSerialModeOrFastHandoff(DMPresource* r) {
  // REQUIRES: the handoff spinlock is held
  DMP_ASSERT(DMPMAP->handoffSpinlock);
  DMP_ASSERT(DMPMAP->state != RunSerial);
  // Don't speculate across this barrier.
  __sync_synchronize();
  DMPMAP->blockedOn = r;
  internal_waitForSerialMode(1);
}

void DMP_fastHandoff(DmpThreadInfo* next, DMPresource* r) {
  // Fast handoff protocol:
  //
  //      Receiver                           Sender
  //
  // 1 if (r.owner != me)                  r.owner = next
  // 2   lock(me.handoffSpinlock)          lock(next.handoffSpinlock)
  // 3   if (r.owner != me)                if (next.blockedOn == r)
  // 4      me.blockedOn = r                 next.blockedOn = NULL
  // 5      me.state = WaitForSerial         next.state = RunParallel
  // 6      unlock(me.handoffSpinlock)     unlock(next.handoffSpinlock)
  // 7      wait
  // 8   else
  // 9      unlock(me.handoffSpinlock)
  //
  // The handoff spinlock is used to atomically check-and-sleep (in the
  // receiver) and check-and-wait (in the sender).
  //
  // This protocol is only meaningful in parallel mode.  The receiver
  // uses double-checked locking to avoid a lock in the common case.
  // This is safe because we insert memory barriers before line 4 in
  // the receiver and before line 2 in the sender.  TODO: is that right?
  //
  // When the receiver awakes (after line 9), the invariant is that either
  // (a) it is running in serial mode, or (b) it is running in parallel
  // mode and has ownership of 'r'.
  //
  // NOTE: the sender should pass ownership before calling this function.
  //
  if (DMPMAP->state == RunSerial)
    return;
  // Wake 'next' if it is waiting for the resource.
  __sync_synchronize();
  DMP_SPINLOCK_LOCK(&next->handoffSpinlock);
  {
    if (next->blockedOn == r) {
#ifdef DMP_ENABLE_FAST_HANDOFF_QUANTUM_OPT
      // Account for the time 'next' spent waiting.
      if (next->schedulingChunk > DMPMAP->schedulingChunk)
        next->schedulingChunk = DMPMAP->schedulingChunk;
#endif
      // Wake.
      DMP_ASSERT(next->state == WaitForSerial);
      next->blockedOn = NULL;
      DMP_setState(next, RunOwnership);
      __sync_add_and_fetch(&DMPserialBarrier, 1);
    }
  }
  DMP_SPINLOCK_UNLOCK(&next->handoffSpinlock);
}

#endif  // DMP_ENABLE_FAST_HANDOFF

//--------------------------------------------------------------
// Kendo-style wait-for-turn
//--------------------------------------------------------------

#ifndef DMP_ENABLE_WB_HBSYNC

void DMP_waitForTurnInRound();

#else

static __thread int64_t TotalChunkDiff;
static __thread int64_t TotalDiffs;

__attribute__((noinline))
void DMP_waitForTurnInRound() {
  // Nop if only one thread is running.
  if (DMPMAP->state == RunSerial || DMPnumRunnableThreads == 1)
    return;

  DMP_ASSERT(DMPMAP->state == RunBuffered);

  const int myChunk = DMPMAP->schedulingChunk;
  const int myID = DMPMAP->threadID;

  // Make sure this is up-to-date to prevent deadlock
  DMPglobalSchedulingChunks[myID].val = myChunk;

//  fprintf(stderr, "CHECK @%llu Me:(%d/%d)\n", DMProundNumber, myID, myChunk);

  // Wait until I am the thread with the fewest ticks so far this quantum.
  for (DmpThreadInfo* dmp = DMPMAP->nextRunnable; dmp != DMPMAP; dmp = dmp->nextRunnable) {
    const int theirID = dmp->threadID;

    // Has the thread passed me?
    // NB: 'state' goes up as quantum-time goes forwards
    // NB: 'schedulingChunk' goes down as time goes forwards

    // NB: the thread may not have started its quantum yet!
    DmpThreadState theirState = VOLATILE_LOAD(dmp->state);
    if (theirState > RunBuffered)
      continue;

    const int theirChunk  = VOLATILE_LOAD(dmp->schedulingChunk);
    const int targetChunk = (theirID < myID) ? myChunk - 1 : myChunk;

    if (theirState >= RunBuffered && theirChunk <= targetChunk)
      continue;

    // Stats
#if 0
    const int diff = theirChunk - targetChunk;
    if (diff >= 0) {
      TotalChunkDiff += diff;
      if (++TotalDiffs == 1000) {
        fprintf(stderr, "AvgDiffs T:%d avg:%lld (this:%d) quantum:%d\n", myID, TotalChunkDiff / TotalDiffs, diff, DMP_SCHEDULING_CHUNK_SIZE);
        TotalChunkDiff = 0;
        TotalDiffs = 0;
      }
    }
#endif

    // Spin until this thread passes me.
    // To reduce cache ping-ponging between us and DMPcommit(), we write a
    // note in (*dmp) then spin on a global field, which is updated rarely.

    int lastVal = theirChunk;

    // NB: Wakeup corner case -- if we started the round before the other thread,
    //     their current chunk may be for the prior round (e.g. if can be < 0),
    //     so fudge 'lastVal' to avoid deadlocks.
    if (theirState < RunBuffered) {
      lastVal = DMP_SCHEDULING_CHUNK_SIZE + 1;
    }

    VOLATILE_STORE(dmp->triggerGlobalSchedulingChunkUpdate, targetChunk);

//    fprintf(stderr, "SPIN! @%llu Me:(%d/%d) Them:(%d/%d)\n", DMProundNumber, myID, myChunk, theirID, theirChunk);

    while (true) {
      // Wait for a signal.
      int nspin = 0;
      int newChunk;
      while ((newChunk = VOLATILE_LOAD(DMPglobalSchedulingChunks[theirID].val)) >= lastVal) {
        if (++nspin = 128) {
          YIELD();
          nspin = 0;
        }
      }

      // Has the thread passed me?
      if (newChunk <= targetChunk)
        break;

      // Not yet; maybe some other thread asked for an earlier wakeup.
      lastVal = newChunk;
      VOLATILE_STORE(dmp->triggerGlobalSchedulingChunkUpdate, targetChunk);
//      fprintf(stderr, "SPIN! @%llu Me:(%d/%d) Them:(%d/%d)\n", DMProundNumber, myID, myChunk, theirID, newChunk);
    }

//    if (nspin >= 10000)
//      fprintf(stderr, "OKAY! @%llu Me:(%d/%d)\n", DMProundNumber, myID, myChunk);
  }

  // NB: another thread could have acquired the mutex we're in line for.
  __sync_synchronize();
}

#endif  // DMP_ENABLE_WB_HBSYNC
