/** Billiards Simulation top level clas  -*- C++ -*-
 * @file
 * @section License
 *
 * Galois, a framework to exploit amorphous data-parallelism in irregular
 * programs.
 *
 * Copyright (C) 2011, The University of Texas at Austin. All rights reserved.
 * UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING THIS
 * SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR ANY PARTICULAR PURPOSE, NON-INFRINGEMENT AND WARRANTIES OF
 * PERFORMANCE, AND ANY WARRANTY THAT MIGHT OTHERWISE ARISE FROM COURSE OF
 * DEALING OR USAGE OF TRADE.  NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH
 * RESPECT TO THE USE OF THE SOFTWARE OR DOCUMENTATION. Under no circumstances
 * shall University be liable for incidental, special, indirect, direct or
 * consequential damages or loss of profits, interruption of business, or
 * related expenses which may arise from use of Software or Documentation,
 * including but not limited to those resulting from defects in Software and/or
 * Documentation, or loss or inaccuracy of data of any kind.
 *
 * @section Description
 *
 * Billiards Simulation top level clas.
 *
 * @author <ahassaan@ices.utexas.edu>
 */


#ifndef _BILLIARDS_H_
#define _BILLIARDS_H_


#include <string>
#include <vector>
#include <sstream>
#include <iostream>
#include <queue>

#include "Galois/Timer.h"
#include "Galois/Statistic.h"
#include "Galois/Galois.h"
#include "Galois/Accumulator.h"
#include "Galois/Runtime/Sampling.h"
#include "llvm/Support/CommandLine.h"
#include "Lonestar/BoilerPlate.h"

#include "TableSectored.h"

namespace cll = llvm::cl;

static const char* const name = "Billiards Simulation";
static const char* const desc = "Simulates elastic collisions between billiard balls on a table";
static const char* const url = "billiards";

static cll::opt<unsigned>   sectorSize("secsize", cll::desc("size of a (square) sector"), cll::init(100));
static cll::opt<unsigned>   xSectors("xsec", cll::desc("number of sectors along length"), cll::init(1));
static cll::opt<unsigned>   ySectors("ysec", cll::desc("number of sectors along height"), cll::init(1));
static cll::opt<unsigned> numballs("balls", cll::desc("number of balls on the table"), cll::init(100));
static cll::opt<unsigned>   endtime("end", cll::desc("simulation end time"), cll::init(100));

static cll::opt<bool> runFlat("flat", cll::desc("Run simulation without introducing sectors/partitions"), cll::init(false));

static cll::opt<bool> veriFlat ("vflat", cll::desc ("Verify against serial flat simulation"), cll::init (false));


template <typename Derived, bool isParallel=true>
class Billiards {

public:
  typedef Galois::GAccumulator<size_t> Accumulator;

  virtual const std::string version () const = 0;
  //! @return number of events processed
  template <typename Tbl_t>
  size_t runSim (Tbl_t& table, std::vector<Event>& initEvents, const FP& endtime, bool enablePrints=false, bool logEvents=false) {
    GALOIS_DIE ("Derived classes must provide the implementation");
  }

private:
  template <typename Tbl_t>
  void verify (const Tbl_t& initial, Tbl_t& final, size_t numEvents, const FP& endtime);

// #define CUSTOM_TESTS
#ifdef CUSTOM_TESTS

  void testA (std::vector<Ball>& balls) {

    sectorSize = 1;
    xSectors = 2;
    ySectors = 1;
    endtime = 10.0;

    balls.clear ();

    balls.emplace_back (0, Vec2 (0.5, 0.5), Vec2 (1.0, 1.0), 1.0, 0.25);
  }

  void testB (std::vector<Ball>& balls) {

    sectorSize = 1;
    xSectors = 2;
    ySectors = 1;
    endtime = 10.0;

    balls.clear ();

    balls.emplace_back (0, Vec2 (0.5, 0.5), Vec2 (1.0, 1.0), 1.0, 0.25);
    balls.emplace_back (1, Vec2 (1.5, 0.5), Vec2 (-1.0, 0.0), 1.0, 0.25);
  }

  void testC (std::vector<Ball>& balls) {

    sectorSize = 1;
    xSectors = 2;
    ySectors = 2;
    endtime = 10.0;

    balls.clear ();

    balls.emplace_back (0, Vec2 (0.5, 0.5), Vec2 (1.0, 1.0), 1.0, 0.25);
    // balls.emplace_back (1, Vec2 (1.5, 0.5), Vec2 (-1.0, 0.0), 1.0, 0.25);
  }

  void testD (std::vector<Ball>& balls) {

    sectorSize = 1;
    xSectors = 2;
    ySectors = 2;
    endtime = 10.0;

    balls.clear ();

    balls.emplace_back (0, Vec2 (0.5, 0.5), Vec2 (1.0, -1.0), 1.0, 0.25);
    // balls.emplace_back (1, Vec2 (1.5, 0.5), Vec2 (-1.0, 0.0), 1.0, 0.25);
  }

  void testE (std::vector<Ball>& balls) {

    sectorSize = 1;
    xSectors = 2;
    ySectors = 2;
    endtime = 10.0;

    balls.clear ();

    balls.emplace_back (0, Vec2 (0.5, 0.5), Vec2 (1.0, 1.0), 1.0, 0.25);
    balls.emplace_back (1, Vec2 (1.5, 1.5), Vec2 (-1.0, -1.0), 1.0, 0.25);
  }

  void testF (std::vector<Ball>& balls) {

    sectorSize = 10;
    xSectors = 3;
    ySectors = 3;
    endtime = 100.0;

    balls.clear ();

    balls.emplace_back (0, Vec2 (0.5, 0.5), Vec2 (1.0, 1.0), 1.0, 0.25);
    balls.emplace_back (1, Vec2 (1.5, 1.5), Vec2 (-1.0, -1.0), 1.0, 0.25);
    balls.emplace_back (2, Vec2 (2.5, 2.5), Vec2 (-1.0, -1.0), 1.0, 0.25);
  }
#endif // CUSTOM_TESTS


  template <typename Tbl_t>
  void runImpl (void) {

#ifdef CUSTOM_TESTS

    std::vector<Ball> balls;

    testF (balls);

    Tbl_t table (balls.begin (), balls.end (), sectorSize, xSectors, ySectors);


#else // CUSTOM_TESTS
    Tbl_t table (numballs, sectorSize, xSectors, ySectors);
#endif // CUSTOM_TESTS

    Tbl_t verCopy (table);

    bool enablePrints = true;
    bool logEvents = true;


    if (enablePrints) {
      table.printState (std::cout);
    }

    std::vector<Event> initEvents;
    table.genInitialEvents (initEvents, unsigned (endtime));

    std::cout << "Number of initial events = " << initEvents.size () << std::endl;

    Galois::StatTimer timer ("Simulation time: ");

    timer.start ();
    Galois::Runtime::beginSampling ();
    size_t numEvents = static_cast<Derived*> (this)->runSim (table, initEvents, unsigned (endtime), enablePrints, logEvents);
    Galois::Runtime::endSampling ();
    timer.stop ();

    std::cout << "Billiards " << version () << ", number of events processed=" << numEvents << std::endl;

    if (!skipVerify) {
      if (veriFlat) {
        verify (static_cast<const Table&> (verCopy), static_cast<Table&> (table), numEvents, unsigned (endtime));

      } else {
        verify (verCopy, table, numEvents, unsigned (endtime));

      }
    }

  }

public:
  void run (int argc, char* argv[]) {
    
    Galois::StatManager sm;
    LonestarStart (argc, argv, name, desc, url);

    if (runFlat) {
      runImpl<Table> ();

    } else {
      runImpl<TableSectored> ();

    }
  }

};


class BilliardsSerialPQ: public Billiards<BilliardsSerialPQ, false> {

  typedef std::priority_queue<Event, std::vector<Event>, Event::ReverseComparator> PriorityQueue;
public:

  virtual const std::string version () const { return "Serial Ordered with Priority Queue"; }

  template <typename Tbl_t>
  GALOIS_ATTRIBUTE_PROF_NOINLINE static void processEvent (Event& e, Tbl_t& table, std::vector<Event>& addList, const FP& endtime) {

      addList.clear ();
      e.simulate ();
      table.addNextEvents (e, addList, endtime);
  }

  template <typename Tbl_t>
  size_t runSim (Tbl_t& table, std::vector<Event>& initEvents, const FP& endtime, bool enablePrints=false, bool logEvents=false) {

    std::printf ("BilliardsSerialPQ: number of initial events: %zd\n", initEvents.size ());

    PriorityQueue pq;

    for (std::vector<Event>::iterator i = initEvents.begin (), ei = initEvents.end ();
        i != ei; ++i) {

        pq.push (*i);
    }

    size_t iter = 0;
    std::vector<Event> addList;

    while (!pq.empty ()) {

      Event e = pq.top ();
      pq.pop ();

      if (enablePrints) {
        std::cout << "Processing event=" << e.str () << std::endl;
      }

      processEvent (e, table, addList, endtime);

      if (logEvents) {
        table.logCollisionEvent (e);
      }

      for (std::vector<Event>::iterator i = addList.begin (), ei = addList.end ();
          i != ei; ++i) {

        pq.push (*i);

        if (enablePrints) {
          std::cout << "Adding event=" << i->str () << std::endl;
        }
      }

      if (enablePrints) {
        table.printState (std::cout);
      }

      ++iter;
    }

    return iter;

  }
};


template <typename Derived, bool isParallel>
template <typename Tbl_t>
void Billiards<Derived, isParallel>::verify (const Tbl_t& initial, Tbl_t& final, size_t numEvents, const FP& endtime) {

  FP initEnergy = initial.sumEnergy ();
  FP finalEnergy = final.sumEnergy ();

  FPutils::checkError (initEnergy, finalEnergy);


  BilliardsSerialPQ serial;
  Tbl_t serialTable(initial);

  std::vector<Event> initEvents;
  serialTable.genInitialEvents (initEvents, endtime);

  Galois::StatTimer timer ("Verfication time (Serial PQ simulation)= ");
  
  timer.start ();
  size_t serEvents = serial.runSim(serialTable, initEvents, endtime, false, true);
  timer.stop ();

  std::cout << "Serial ordered numEvents=" << serEvents << std::endl;

  if (serEvents != numEvents) {
    std::cerr << "WARNING: number of events differ from verification run, normal=" 
      << numEvents <<  " != verification=" << serEvents << std::endl;
  }

  FP serEnergy = serialTable.sumEnergy ();

  FPutils::checkError (serEnergy, initEnergy);
  FPutils::checkError (serEnergy, finalEnergy);


  //advance both tables to same endtime
  final.advance (endtime);
  serialTable.advance (endtime);

  // check the end state of both simulations
  // pass true to print differences 
  bool result = serialTable.template cmpState<true> (final);

  if (!result) {
    std::cerr << "ERROR, Comparison against serial ordered simulation failed due to above differences" << std::endl;

    // std::cout << "<<<<<<<<<<< Event logs from Original Run >>>>>>>>>>>>>" << std::endl;
    // final.printEventLogs ();
// 
    // std::cout << "<<<<<<<<<<< Event logs from Verification Run >>>>>>>>>>>>>" << std::endl;
    // serialTable.printEventLogs ();

    final.diffEventLogs (serialTable, "Serial Flat Run");

    abort ();

  } else {
    std::cout << "OK, Result verified against serial ordered simulation" << std::endl;
  }


}

#endif // _BILLIARDS_H_

