/** Scalable local worklists -*- C++ -*-
 * @file
 * @section License
 *
 * This file is part of Galois.  Galoisis a gramework to exploit
 * amorphous data-parallelism in irregular programs.
 *
 * Galois is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Galois is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Galois.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * @section Copyright
 *
 * Copyright (C) 2015, The University of Texas at Austin. All rights
 * reserved.
 *
 * @author Andrew Lenharth <andrewl@lenharth.org>
 */
#ifndef GALOIS_RUNTIME_WORKLISTALT_H
#define GALOIS_RUNTIME_WORKLISTALT_H

#include "Galois/WorkList/Simple.h"
#include "Galois/Substrate/CompilerSpecific.h"

namespace Galois {
namespace WorkList {

template<typename QueueTy>
Galois::optional<typename QueueTy::value_type>
stealHalfInPackage(Substrate::PerThreadStorage<QueueTy>& queues) {
  unsigned id = Substrate::ThreadPool::getTID();
  unsigned pkg = Substrate::ThreadPool::getPackage();
  unsigned num = Galois::getActiveThreads();
  QueueTy* me = queues.getLocal();
  Galois::optional<typename QueueTy::value_type> retval;
  
  //steal from this package
  //Having 2 loops avoids a modulo, though this is a slow path anyway
  auto& tp = Substrate::getSystemThreadPool();
  for (unsigned i = id + 1; i < num; ++i)
    if (tp.getPackage(i) == pkg)
      if ((retval = me->steal(*queues.getRemote(i), true, true)))
	return retval;
  for (unsigned i = 0; i < id; ++i)
    if (tp.getPackage(i) == pkg)
      if ((retval = me->steal(*queues.getRemote(i), true, true)))
	return retval;
  return retval;
}

template<typename QueueTy>
Galois::optional<typename QueueTy::value_type>
stealRemote(Substrate::PerThreadStorage<QueueTy>& queues) {
  unsigned id = Substrate::ThreadPool::getTID();
  //  unsigned pkg = Runtime::LL::getPackageForThread(id);
  unsigned num = Galois::getActiveThreads();
  QueueTy* me = queues.getLocal();
  Galois::optional<typename QueueTy::value_type> retval;
  
  //steal from this package
  //Having 2 loops avoids a modulo, though this is a slow path anyway
  for (unsigned i = id + 1; i < num; ++i)
    if ((retval = me->steal(*queues.getRemote(i), true, true)))
      return retval;
  for (unsigned i = 0; i < id; ++i)
    if ((retval = me->steal(*queues.getRemote(i), true, true)))
      return retval;
  return retval;
}

template<typename QueueTy>
class PerThreadQueues : private boost::noncopyable {
public:
  typedef typename QueueTy::value_type value_type;
  
private:
  Substrate::PerThreadStorage<QueueTy> local;

  Galois::optional<value_type> doSteal() {
    Galois::optional<value_type> retval = stealHalfInPackage(local);
    if (retval)
      return retval;
    return stealRemote(local);
  }

  template<typename Iter>
  void fill_work_l2(Iter& b, Iter& e) {
    unsigned int a = Galois::getActiveThreads();
    unsigned int id = Substrate::ThreadPool::getTID();
    unsigned dist = std::distance(b, e);
    unsigned num = (dist + a - 1) / a; //round up
    unsigned int A = std::min(num * id, dist);
    unsigned int B = std::min(num * (id + 1), dist);
    e = b;
    std::advance(b, A);
    std::advance(e, B);
  }

  // Runtime::LL::SimpleLock<true> L;
  // std::vector<unsigned> sum;

  template<typename Iter>
  void fill_work_l1(Iter b, Iter e) {
    Iter b2 = b;
    Iter e2 = e;
    fill_work_l2(b2, e2);
    unsigned int a = Galois::getActiveThreads();
    //    unsigned int id = Runtime::LL::getTID();
    std::vector<std::vector<value_type> > ranges;
    ranges.resize(a);
    while (b2 != e2) {
      unsigned i = getID(*b2);
      ranges[i].push_back(*b2);
      ++b2;
      if (ranges[i].size() > 128) {
	local.getRemote(i)->push(ranges[i].begin(), ranges[i].end());
	ranges[i].clear();
      }
    }
    // L.lock();
    // if (sum.empty())
    //   sum.resize(a + 1);
    // sum[a]++;
    // std::cerr << id << ":";
    // for (unsigned int x = 0; x < a; ++x) {
    //   std::cerr << " " << ranges[x].size();
    //   sum[x] += ranges[x].size();
    // }
    // std::cerr << "\n";
    // if (sum[a] == a) {
    //   std::cerr << "total:";
    //   for (unsigned int x = 0; x < a; ++x)
    // 	std::cerr << " " << sum[x];
    //   std::cerr << "\n";
    // }
    // L.unlock();
    for (unsigned int x = 0; x < a; ++x)
      if (!ranges[x].empty())
	local.getRemote(x)->push(ranges[x].begin(), ranges[x].end());
  }

public:
  template<typename Tnew>
  using retype = PerThreadQueues<typename QueueTy::template retype<Tnew>::type>;

  template<bool newConcurrent>
  using rethread = PerThreadQueues<typename QueueTy::template rethread<newConcurrent>::type>;

  void push(const value_type& val) {
    local.getLocal()->push(val);
  }

  template<typename Iter>
  void push(Iter b, Iter e) {
    local.getLocal()->push(b,e);
  }

  template<typename RangeTy>
  void push_initial(RangeTy range) {
    fill_work_l1(range.begin(), range.end());
  }

  Galois::optional<value_type> pop() {
    Galois::optional<value_type> retval = local.getLocal()->pop();
    if (retval)
      return retval;
    return doSteal();// stealHalfInPackage(local);
  }
};
//GALOIS_WLCOMPILECHECK(LocalQueues);

template<typename WLTy = GFIFO<int>, typename T = int>
class LocalWorklist : private boost::noncopyable {
  typedef typename WLTy::template rethread<false> lWLTy;
  Substrate::PerThreadStorage<lWLTy> local;

public:
  template<bool newconcurrent>
  using rethread = LocalWorklist<typename WLTy::template rethread<newconcurrent>, T>;

  template<typename Tnew>
  using retype = LocalWorklist<typename WLTy::template retype<Tnew>, Tnew>;

  typedef T value_type;

  void push(const value_type& val) {
    local.getLocal()->push(val);
  }

  template<typename Iter>
  void push(Iter b, Iter e) {
    local.getLocal()->push(b, e);
  }

  template<typename RangeTy>
  void push_initial(RangeTy range) {
    local.getLocal()->push(range.local_begin(), range.local_end());
  }

  Galois::optional<value_type> pop() {
    return local.getLocal()->pop();
  }
};
GALOIS_WLCOMPILECHECK(LocalWorklist)

template<typename T, typename OwnerFn, template<typename, bool> class QT, bool distributed = false, bool isStack = false, int chunksize=64, bool concurrent=true>
class OwnerComputeChunkedMaster : private boost::noncopyable {
  class Chunk : public Galois::FixedSizeRing<T, chunksize>, public QT<Chunk, concurrent>::ListNode {};

  Runtime::FixedSizeHeap heap;
  OwnerFn Fn;

  struct p {
    Chunk* cur;
    Chunk* next;
  };

  typedef QT<Chunk, concurrent> LevelItem;

  Substrate::PerThreadStorage<p> data;
  detail::squeue<distributed, Substrate::PerPackageStorage, LevelItem> Q;

  Chunk* mkChunk() {
    return new (heap.allocate(sizeof(Chunk))) Chunk();
  }
  
  void delChunk(Chunk* C) {
    C->~Chunk();
    heap.deallocate(C);
  }

  void pushChunk(Chunk* C)  {
    unsigned int tid = Substrate::ThreadPool::getTID();
    unsigned int index = isStack ? Fn(C->back()) : Fn(C->front());
    if (tid == index) {
      LevelItem& I = Q.get();
      I.push(C);
    } else {
      unsigned int mindex = Substrate::getSystemThreadPool().getPackage(index);
      LevelItem& I = Q.get(mindex);
      I.push(C);
    }
  }

  Chunk* popChunkByID(unsigned int i)  {
    LevelItem& I = Q.get(i);
    return I.pop();
  }

  Chunk* popChunk()  {
    int id = Q.myEffectiveID();
    Chunk* r = popChunkByID(id);
    if (r)
      return r;

    for (int i = id + 1; i < (int) Q.size(); ++i) {
      r = popChunkByID(i);
      if (r) 
	return r;
    }

    for (int i = 0; i < id; ++i) {
      r = popChunkByID(i);
      if (r)
	return r;
    }

    return 0;
  }

  T* pushi(const T& val, p* n)  {
    T* retval = 0;

    if (n->next && (retval = n->next->push_back(val)))
      return retval;
    if (n->next)
      pushChunk(n->next);
    n->next = mkChunk();
    retval = n->next->push_back(val);
    assert(retval);
    return retval;
  }

public:
  typedef T value_type;

  template<bool newconcurrent>
  using rethread = OwnerComputeChunkedMaster<T, OwnerFn,QT, distributed, isStack, chunksize, newconcurrent>;

  template<typename Tnew>
  using retype = OwnerComputeChunkedMaster<Tnew, OwnerFn, QT, distributed, isStack, chunksize, concurrent>;

  OwnerComputeChunkedMaster() : heap(sizeof(Chunk)) { }

  void flush() {
    p& n = *data.getLocal();
    if (n.next)
      pushChunk(n.next);
    n.next = 0;
  }
  
  //! Most worklists have void return value for push. This push returns address
  //! of placed item to facilitate some internal runtime uses. The address is
  //! generally not safe to use in the presence of concurrent pops.
  value_type* push(const value_type& val)  {
    p* n = data.getLocal();
    return pushi(val, n);
  }

  template<typename Iter>
  void push(Iter b, Iter e) {
    p* n = data.getLocal();
    while (b != e)
      pushi(*b++, n);
  }

  template<typename RangeTy>
  void push_initial(RangeTy range) {
    push(range.local_begin(), range.local_end());
  }

  Galois::optional<value_type> pop()  {
    p& n = *data.getLocal();
    Galois::optional<value_type> retval;
    if (isStack) {
      if (n.next && (retval = n.next->extract_back()))
	return retval;
      if (n.next)
	delChunk(n.next);
      n.next = popChunk();
      if (n.next)
	return n.next->extract_back();
      return Galois::optional<value_type>();
    } else {
      if (n.cur && (retval = n.cur->extract_front()))
	return retval;
      if (n.cur)
	delChunk(n.cur);
      n.cur = popChunk();
      if (!n.cur) {
	n.cur = n.next;
	n.next = 0;
      }
      if (n.cur)
	return n.cur->extract_front();
      return Galois::optional<value_type>();
    }
  }
};

template<typename OwnerFn=DummyIndexer<int> , int chunksize=64, typename T = int, bool concurrent=true>
class OwnerComputeChunkedLIFO : public OwnerComputeChunkedMaster<T,OwnerFn,ConExtLinkedQueue, true, true, chunksize, concurrent> {};
GALOIS_WLCOMPILECHECK(OwnerComputeChunkedLIFO)


}//End namespace
} // end namespace Galois

#endif
